#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

RECLASS_NAME = 'reclass'
DESCRIPTION = ('merge data by recursive descent down an ancestry hierarchy '
               '(forked extended version)')
VERSION = '1.6.0'
AUTHOR = 'martin f. krafft / Andrew Pickford / salt-formulas community'
AUTHOR_EMAIL = 'salt-formulas@freelists.org'
MAINTAINER = 'salt-formulas community'
MAINTAINER_EMAIL = 'salt-formulas@freelists.org'
COPYRIGHT = ('Copyright © 2007–14 martin f. krafft, extensions © 2017 Andrew'
             ' Pickford, extensions © salt-formulas community')
LICENCE = 'Artistic Licence 2.0'
URL = 'https://github.com/salt-formulas/reclass'
